module.exports = {
  baseUrl: '/f3d',
  work: './build-tmp',
  examples: [],
  config: {
    title: 'F3D',
    description: '"A fast and minimalist 3D viewer"',
    subtitle: '"A fast and minimalist 3D viewer"',
    author: 'Kitware SAS, Michael Migliore, Mathieu Westphal',
    timezone: 'CET',
    url: 'https://f3d-app.github.io/f3d/',
    root: '/f3d/',
    github: 'f3d-app/f3d',
    google_analytics: 'G-QQJSFRXCT0',
  },
  copy: [],
};
